'+----------------------------------------------------+
'|     logic analyzer framework research program      |
'|                                                    |
'| Volhout                                 20-1-2023  |
'+----------------------------------------------------+
'samples GP0...GP15 with triggering on GP22
'uses PIO and ring buffer under DMA.
'requires V50707b8 or newer (function names changed)
  
'---------------------- version control ----------------
' LA_3  uses DMA after manual trigger, serial output
' LA_4  uses DMA and ring buffer, serial output
' LA_5  testbed to debug DMA
' LA_6  restructure (this version) tbd PIO is ready
' LA_7  unroll circular buffer to linear array
' LA_8  triggering works
' LA_9  basic VGA display added
' LA_10 pan/scan added on cursor keys
' LA_11 8 chan, bar, sampling implemented
' LA_12 6/8 channel, trigger overhaul.
' LA_13 single/run
  
  
'--------------  prepare and init the picomite for LA ---------------
  
  max_chan_no%=5                      'number of input channels
  samples% = 4096                     'must be 2^x, larger than 256
  st%=0:sp%=samples%-1                'whole buffer shown default
  frequency_index%=3                  '10kHz deafult sampling speed
  single%=&h7fffffff                  'large, anything > 0 for repreat
  
'  generate_test_signal                '--debug-- test signal
  
  PIO_prog                            'program the PIO to sample
  
  init_memory                         'set up data buffers
  init_trigger_config                 'set up choice arrays
  init_sampling_rate                  'different sampling rates
  
  set_screen                          'set default screen parameters
  
  
'------------------------------ main loop -------------------------------
  
  Do
    
    If aa$="" Then
      aa$=Inkey$
    EndIf
    
'   read keyboard
'   if "<" then move centre, display
'   if ">" then move centre, display
'   if "^" then zoom in, display
'   if "V" then zoom out, display
'   if "t" then unarm, goto trigger setup
'   if "+/-" then unarm, change sample setup
'   if number then change channel name
'   if PIO ready then update, display
'loop
    
'the old linear user interface
' Do
    
'decode trigger
    If aa$="t" Then
      Print @(0,0);"set trigger (x,1,0,/,\), type 1 char per GP pin";
      Input z$
      If Len(z$)=(1+max_chan_no%) Then
        setup_trigger_pattern
      EndIf
    EndIf
    
'pan/scan on cursor keys
    If aa$=Chr$(128) Then
      cp%=(sp%+st%)/2:rn%=(sp%-st%)
      rn%=Max(rn%/4,64)
      st%=cp%-rn%:sp%=cp%+rn%
    EndIf
    If aa$=Chr$(129) Then
      cp%=(sp%+st%)/2:rn%=(sp%-st%)
      st%=Max(cp%-rn%,0):sp%=Min(cp%+rn%,samples%-1)
    EndIf
    If aa$=Chr$(130) Then
      cp%=(sp%+st%)/2:rn%=(sp%-st%)/2
      cp%=cp%-rn%/2
      st%=Max(cp%-rn%,0):sp%=Min(cp%+rn%,samples%-1)
    EndIf
    If aa$=Chr$(131) Then
      cp%=(sp%+st%)/2:rn%=(sp%-st%)/2
      cp%=cp%+rn%/2
      st%=Max(cp%-rn%,0):sp%=Min(cp%+rn%,samples%-1)
    EndIf
    
    If aa$="+" Then
      frequency_index%=Min(frequency_index%+1,13)
      f0=sf%(frequency_index%)
      Box 56,42,52,16,,0,1             'full buffer
      Print @(62,44) Right$("  "+sr$(frequency_index%),4);"S"
      single%=single%+1                'another sample run
    EndIf
    If aa$="-" Then
      frequency_index%=Max(frequency_index%-1,0)
      f0=sf%(frequency_index%)
      Box 56,42,52,16,,0,1             'full buffer
      Print @(62,44) Right$("  "+sr$(frequency_index%),4);"S"
      single%=single%+1                'another sample run
    EndIf
    
'run/stop control
    if aa$="s" then
      single%=1
      print @(10,44)"SNGL"
    end if
    if aa$="r" then
      single%=&h7fffffff
      print @(10,44)" RUN"
    end if
    
'change trigger with numeric keys
    if asc(aa$)>=&h30 and asc(aa$) =< (&h30 + max_chan_no%) then
      cha%=val(aa$)
      inc tr%(cha%),1
      if tr%(cha%)=5 then tr%(cha%)=0
      trigger_legend
      trigger_create_program
    end if
    
'end of key decoding
    aa$=""
    
'print zoom banner
    show_zoom
    if single% > 0 then
'start la, wait for trigger
      arm_la
      
'calculate minimal circular buffer prefill time
      filltime%=Int(1000*3*(samples%-posttrigger%)/f0)+10
      Pause filltime%     'fill pre-trigger buffer
      
'generate_trigger
      enable_trigger
      
      
'wait for trigger to happen
      Print @(350,45) "ARM"
      Do
        aa$=Inkey$
      Loop While Pin(gp22)=1 And aa$=""
      Print @(350,45) "TRIG"
      
'the PIO should autocomplete, wait for post trigger length% written
      filltime%=Int(1000*3*posttrigger%/f0)+10
      Pause filltime%    'fill post-trigger buffer
      
'stop DMA, this will call ReadyInt that stops the PIO and re-init it.
      PIO DMA RX OFF                           'also unpacks the data
      
'signal that the acquisition has finished
      single%=single%-1
      if single%=0 then print @(10,44)"HALT"
    end if
    
'display the data
    scrnprint st%,sp%
  Loop
  
End
  
  
  
  
'-----------------------------------SUBS MMBasic ------------------------------
  
'convert the trigger string to PIO instructions for state machine 1
'program these instructions into pre-defined locations
Sub setup_trigger_pattern
  For h%=0 To max_chan_no%
    For g%=0 To 4
      If Mid$(z$,h%+1,1)=c$(g%) Then
        tr%(h%)=g%
        trigger_legend
        trigger_create_program
'if h%<6 then
'PIO program line 1,14+3*h%,p%(g%)     'program locations 14,17,....29
'end if
      EndIf
    Next g%
  Next h%
End Sub
  
  
'start the trigger engine
Sub enable_trigger
  PIO init machine 1,1,f1,p1,e1,s1,10   'start address = 10 (0x0A)
  PIO start 1,1
End Sub
  
  
'trigger characters related to PIO instructions
Sub init_trigger_config
  Dim c$(4) length 2 = ("x","1","0","/","\")  'text for the 5trigger  modes
  Dim tr%(max_chan_no%)                       'trigger mode default 0
  
'the instructions must be re-done in dynamic programming
  Dim p%(4)=(&hE020,&h002A,&h004A,&hE020,&hE020)
End Sub
  
  
'relation between PIO clock and sampling rates
Sub init_sampling_rate
  Dim sr$(13) length 4=("1k","2k","5k","10k","20k","50k","100k","200k","500k","1M","2M","5M","10M","21M")
  Dim sf%(13)=(3e3,6e3,15e3,3e4,6e4,15e4,3e5,6e5,15e5,3e6,6e6,15e6,31e6,63e6)
End Sub
  
  
'initializes the screen (colors, font, lines)
Sub set_screen
  MODE 1:Colour 0,RGB(WHITE):CLS :Font 1   'generic screen setup from Martin H
  TILE 0,0,0,RGB(white),40,30              'default all tiles
  dim nam$(max_chan_no%) length 9
  Local f%
  
'indicate initial sampling speed and run mode
  Box 0,42,52,16,,0,1:print @(10,44)" RUN"
  Box 56,42,52,16,,0,1:Print @(62,44) Right$("  "+sr$(frequency_index%),4);"S"
  
'default color pattern and channel names for LA from Martin H
  For f%=0 To max_chan_no%
    TILE 0,5+f%*3,0,RGB(WHITE),7,2:TILE 7,5+f%*3,0,RGB(yellow),33,2
    TILE 0,4+f%*3,0,RGB(midgreen),40,1:Line 0,64+f%*48,640,64+f%*48
    nam$(f%)="Channel "+str$(f%)
  Next f%
  
'posh for 6 channel mode
  TILE 0,22,0,RGB(midgreen),40,1:Line 0,352,640,352
  
'show default startup information fro channels and trigger
  channel_legend
  trigger_legend
  
End Sub
  
  
'shows the trigger settings on screen
Sub trigger_legend
  Local f%,i%
  For f%=0 To max_chan_no%               'for all channels
    Box 30,96+48*f%,72,16,,1,1           'erase all boxes and text
    For i%=0 To 4                        'step through all 5 trigger modes
      If tr%(f%)=i% Then
        Box 30+i%*14,96+48*f%,16,16     'write current trigger boxes
      EndIf
      Print @(34+i%*14,98+48*f%) c$(i%)   're-write text
    Next i%
  Next f%
End Sub
  
  
'shows the channel names on screen
Sub channel_legend
  Local f%
  For f%=0 To max_chan_no%
    Font 2:Print @(0,84+48*f%); f%
    Font 1:Print @(30,82+48*f%); nam$(f%)
  Next f%
End Sub
  
  
'unpack packed%() to unpacked%() such that trigger is in the centre
Sub unroll
  varindex%=(Pio(DMA RX POINTER)-Peek(VARADDR packed%()))/8   'bytes -> 64bits
  beginsamples%=(Pio(DMA RX POINTER)-Peek(VARADDR packed%()))/4  'bytes->32bit

  'in case DMA timeout, the varindex% exceeds array limits, correct it
  if varindex% > length%-1 then
    varindex%=0                'in case timeout bug
    if beginsamples% > samples%-1 then beginsamples%=0        'in case timeout bug
  end if

  endsamples% = samples% - beginsamples%
  Memory unpack Peek(varaddr packed%(varindex%)),unpacked%(),endsamples%,32
  Memory unpack packed%(),Peek(varaddr unpacked%(endsamples%)),beginsamples%,32
End Sub
  
  
'when DMA is ready stop PIO's and unroll buffer
Sub ReadyInt
  PIO STOP 1,0
  PIO stop 1,1
  unroll
  Math set 0,packed%()
End Sub
  
  
'Start the DMA to accept FIFO data from PIO 1.0, and set the posttrigger value
Sub arm_la
'prepare the PIO 1.0 for new DMA
  PIO init machine 1,0,f0,p0,e0,s0,0     'start address = 0
  
'the DMA will time out after &h7FFFFFFF samples, or when stopped
  PIO DMA RX 1,0,&h7FFFFFFF,packed%(),ReadyInt,32,samples% 'start DMA ring buf
  
'write to PIO the number of post trigger samples, starts logging pre-trigger
  PIO WRITE 1,0,1,posttrigger%
End Sub
  
  
Sub init_memory
'define data buffers
'The logic analyzer captures samples% samples into the packed%() array'
'The DMA buffer has 64bit width, packed with 32bit FIFO (2 per cell) = length%
'The size of the ring buffer is specified in bytes, so size is 8 * length%
'the samples will be unpacked into unpacked%() at 32 bits per cell (=2*lenth%)
  
  posttrigger% = samples% / 2        'set trigger in centre of circular buffer
  length% = samples% / 2             'packed buffer = 64bit, data buffer 32bit
  Dim unpacked%(samples%-1)          'array to put the 32 bit samples (base 0)
  
  Dim packed%                        'name of array to become circular buffer
  PIO MAKE RING BUFFER packed%,8*length%
  
End Sub
  
  
' Test signal: Three phase motor drive - picomite V50706+
Sub generate_test_signal
' runs autonomously on GP0...GP5 50Hz
  
  SetPin gp0,pwm  'CH 0a
  SetPin gp1,pwm  'CH 0b
  SetPin gp2,pwm  'CH 1a
  SetPin gp3,pwm  'CH 1b
  SetPin gp4,pwm  'CH 2a
  SetPin gp5,pwm  'CH 2b
  
  PWM 0, 50, 11.11, -88.88, 1, 1
  PWM 1, 49, 33.33, -66.66, 1, 1
  PWM 2, 50, 33.33, -66.66, 1, 1
  
  PWM sync 0, 100/3, 200/3
  Pause 20
End Sub
  
  
'program the PIO with the sampler code (PIO 1.0) and default trigger PIO 1.1
'PIO 1.1 is reprogrammed each time trigger is changed
Sub PIO_prog
  
'assign GP22 (trigger_not) to PIO
  SetPin gp22,pio1
  
'the PIO 1.0 program: the sampler PIO reads GP0..GP7 brute force
'and pushes data into FIFO. The clock speed determines the
'sampling rate. There are 3 instructions per cycle
'after trigger (line 6) is detected "posttrigger%" samples are added to FIFO
  
  
'clean start
  PIO clear 1
  
  
'program pio1.0 : sampler
  PIO program line 1,0,&hE081           'SET GP22 output
  
  PIO program line 1,1,&hE001           'SET GP22 high
  PIO program line 1,2,&h80A0           'PULL block
  PIO program line 1,3,&hA027           'mov(x,osr)
  
  PIO program line 1,4,&h4001+max_chan_no%  'IN pins 6/8
  PIO program line 1,5,&h8000           'PUSH noblock
  PIO program line 1,6,&h00C4           'JMP (GP22) 4
  
  PIO program line 1,7,&h4001+max_chan_no%  'IN pins 6/8
  PIO program line 1,8,&h8000           'PUSH noblock
  PIO program line 1,9,&h0047           'JMP (X<>0) 7 X--
  
  
'initial trigger program: run continuous (set GP22 low)
  PIO program line 1,10,&hE000          'set GP22 low
  PIO program line 1,11,&h000B          'jmp always to itself
  
  
'configuration PIO 1.0
  f0=4e4                             'PIO run at 10kHz
  p0=Pio(pinctrl 0,1,0,gp0,,GP22,)   'IN base = GP0, SET = GP22
  e0=Pio(execctrl gp22,1,9)          'wrap 9 to 1, gp22 is conditional JMP pin
  s0=Pio(shiftctrl 0,0,0,0,0,1)      'shift in through LSB, out is not used
  
  
'configuration PIO 1.1
  f1=63e6                            'PIO run at 63MHz
  p1=Pio(pinctrl 0,1,0,gp0,,GP22,)   'IN base = GP0, SET = GP22
  e1=Pio(execctrl gp0,10,30)         'wrap 30 to 10
  s1=Pio(shiftctrl 0,0,0,0,0,1)      'shift in through LSB, out through LSB
  
  
'write the default configurations
  PIO init machine 1,0,f0,p0,e0,s0,0    'start address = 0
  PIO init machine 1,1,f1,p1,e1,s1,10   'start address = 10 (0x0A)
  
End Sub
  
  
'print the traces on the serial output
Sub scrnprint st%,sp%
'print section st% to sp% of the linear array to serial (screen)
  Local z%,y%,x%,stp%,mask%,one%,eras%,bit%
  stp%=Int(0.5+(sp%-st%)/128)
  Timer =0
  For j%=0 To max_chan_no%
    z%=j%*48+85
    mask%=2^j%
    Box 112,z%-17,524,14,,1,1  'erase
    For i%=st% To sp% Step stp%
      bit%=((unpacked%(i%) And mask%)=0)
      one%=20*bit%
      Eras%=20*(1-bit%)
      x%=116+4*(i%-st%)/stp%
      y%=z%+one%
      Line x%,y%,x%+4,y%,,0
      y%=z%+eras%
      Line x%,y%,x%+4,y%,,1
      If i%=length% And tr%(j%)<>0 Then Print @(x%-4,z%-17) "T"
    Next i%
  Next j%
'Print @(570,0) Timer
End Sub
  
  
'show bar st%-sp% in relation to full range 0-samples%
Sub show_zoom
  Local stp%
  Box 116,42,513,16,,0,1             'full buffer
  stp%=samples%/512
  Box 117+(st%/stp%),44,(sp%-st%)/stp%,12,,0,0
End Sub
  
  
'from the tr%() array create the trigger program for PIO 1.1
Sub trigger_create_program
'when there is an edge trigger, it is placed at start of the program
'after that the logic conditions are added
'there can be only 1 edge trigger in this engine, the first in string
  
  Local trigger_start_address%, trigger_end_address%, i%
  trigger_start_address% = 10            'address &h0A
  trigger_end_address%= 31               'last location PIO program
  write_pio%=trigger_start_address%
  
'check edge trigger, for edge trigger use the PIO WAIT GPx instruction
'a falling edge is WAIT GPx=1, WAIT GPx=0. Rising edge WAIT GPx=0, WAIT GPx=1
  
  For i%=0 To max_chan_no%
    If tr%(i%) > 2 Then
      
'found edge trigger for channel i%
'below code can be optimized further
      
      If tr%(i%)=3 Then     'rising edge
        inc_write_pio &h2000+i%
        inc_write_pio &h2080+i%
      Else                  'falling edge
        inc_write_pio &h2080+i%
        inc_write_pio &h2000+i%
      EndIf
      i%=max_chan_no%                 'search no further, skip the rest
    EndIf
  Next i%
  
  
'check the static conditions. This shifts data in the ISR, copies to OSR
  inc_write_pio &h4001+max_chan_no%   'IN ISR 6/8
  inc_write_pio &hA0E6                'MOV ISR->OSR
  
  
'Each input processed by reading a new bit from OSR and checking it
'edge triggered pin is skipped (first one is processed, rest is ignored)
  
  For i%=0 To max_chan_no%
    Select Case tr%(i%)
      Case 0,3,4                      'x / \ do nothing, shift to Y trash
        inc_write_pio &h6041          'OUT OSR->Y 1
      Case 1                          '1
        inc_write_pio &h6021          'OUT OSR->X 1
        inc_write_pio &h002A          'trigger GPx=1
      Case 2                          '0
        inc_write_pio &h6021          'OUT OSR->X 1
        inc_write_pio &h004A          'trigger GPx=0
    End Select
  Next i%
  
'when all triggers are true, we set GP22 low as trigger for the sampler
  inc_write_pio &hE000                'set GP22 low
  inc_write_pio &h000A                'jmp top (needed when memory not full)
'print @(10,10) write_pio%
  
' maximum program length is:
' without edge trigger: 2 (shift in) + 8*2 (8 channels) + 2 (trigger+jump) = 20
' with edge trigger: 2 (edge) + 2 (shift in) + 7 * 2 (7 remaining channels)
'                                    + 1 (skipped edge) + 2 (trigger+jump) = 21
' start addres = 10, end = 31 -> 21 will always fit. No check needed.
End Sub
  
  
'this programs 1 PIO line and advances to the next line
Sub inc_write_pio instr%
  PIO program line 1,write_pio%,instr%
  Inc write_pio%,1
End Sub
